#!/usr/bin/env ruby
#   SyntenyMapper.rb - Version 1H
#   Generates a CMAP (Contig Mapping) file given a FASTA genome and target file as input.
#
#   Code Copyright (C) 2016 Daniel Veltri (dan.veltri@gmail.com)
#
#   This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation version 3 of the License.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#   SimpleSynteny Copyright (C) 2016 Daniel Veltri
#   This program comes with ABSOLUTELY NO WARRANTY; This is free software, and you are
#   welcome to redistribute it under certain conditions; for details see the LICENSE.txt
#   file that should have been distributed with this script. 
#===============================================
# Created on 9.18.2014
# Last Updated: 4.7.2016
# - Added segment_hits function to allow for repeats of genes rather than just the one with best HSP per-contig
# Update: 9.11.2015
# - Added extra option to print contig number (in order of file) at start of CMAP line for CFinder usage
# Update: 6.2.2015
# - Redid mapping to start at lowest HSP
# - Updated identity cutoff to factor for full-length query by mapping HSP tiles.
#===============================================
# Basic Usage: ruby SyntenyMapper.rb <targets.fasta> <genome.fasta> <base_output_name> [options]
#
# Dependencies: BioRuby GEM, NCBI-BLAST+ (and blastall from legacy BLAST) binaries.
# *WARNING* User needs to adjust user-set variables before running script!
# See the SimpleSynteny Command Line Manual for details.
#===============================================

#==================================================
#REQUIRED LIBRARIES
require 'bio'

#==================================================
#USER-SET GLOBALS 
$LOCAL_BLASTALL_LOC = '/home/.../ncbi-blast/bin/blastall' # path to legacy blastall program
$LOCAL_MAKEBLASTDB_LOC = '/home/.../ncbi-blast/bin/makeblastdb' # path to makeblastdb
$BLAST_VR = '2.2.26' # BLAST version shown on report

$TOKEN = Time.now.to_i # used to name blast db
$CLEAN_DB = true # remove blast db files when done?
$VERSION='1H'

#========================
#HELPER FUNCTIONS

#On Entry: cmap_line is a valid CMAP string
#On Exit: cmap_line is flipped including coordinates and gene directions and returned as string
def flip_contig(cmap_line)
  cstart = 0
  cend = 0
  genes = []
  if cmap_line =~ /(\{)(\d+)(.*?)(\d+)(\})/
    cstart = $2.to_i
    cend = $4.to_i
    genes = $3.split(" ]  [ ")
  end

  pstring = "{#{cstart}  "
  revgenes = genes.reverse
  for i in 0 ... genes.length
    e = revgenes[i]
    if e =~ /(\d+)(\|)(\d+)(.*\).*?)(\d+)(\|)(\d+)/
      gfstart = $1.to_i
      gfend = $7.to_i
      gstart = $3.to_i
      gend = $5.to_i
      gstring = $4
      raise "\n\nWARNING! Flipped gene coords don't match contig length!\n#{e}\nContig Len: #{cend}\nGene Start: #{gfstart} End: #{gfend}\n\n\n" if (cend - gfend + 1) <= 0
      if gstring =~ /(\(\d+\|\d+)(.*)(\s+==>\s+)(\d+\|\d+\))/
        pstring += "[ #{(cend - gfend + 1)}|#{(cend - gend + 1)} #{$1}  <== #{$2.strip}  #{$4} #{(cend - gstart + 1)}|#{(cend - gfstart + 1)} ]  "
      elsif gstring =~ /(\(\d+\|\d+)(\s+<==\s+)(.*\s+)(\d+\|\d+\))/
        pstring += "[ #{(cend - gfend + 1)}|#{(cend - gend + 1)} #{$1}  #{$3.strip} ==>  #{$4} #{(cend - gstart + 1)}|#{(cend - gfstart + 1)} ]  "
      else
        raise "ERROR in flip_contig! Invalid gene format: #{gstring}"
      end
    else
      raise "ERROR in flip_contig! Invalid contig format: #{e}"
    end
  end
  pstring += "#{cend}}\n"
  return pstring
end


#On Entry: fastafile is file name of valid fasta file, target_definition is definition line to retrieve from fastafile and MUST be there else function will raise error!
#On Exit: block is returned with target_definition's sequence, target definitions reverse complement and the sequence length
def get_contig(fastafile, target_definition) 
  count = 0
  Bio::FastaFormat.open(fastafile).each do |f|
    count += 1
    return f.seq, f.naseq.reverse_complement.upcase, f.seq.length if f.definition == target_definition 
  end
  
  #If we reach here it's an error...
  if count == 0
    raise "ERROR in get_contig_length! Empty target fasta file given!"
  else
    raise "ERROR in get_contig_length! Invalid definition '#{target_definition}' given!"
  end
end


#Splits hits into uniq ranges to consider in order to handle target repeats per reviewer request
#Definitely should code a more efficient way to do this given time.
#On entry: hit from BioRuby BLAST report factory is provided for one gene on one contig, evalue_cutoff is threshold under which hits must be to be considered
#On exit: array container contains arrays with start-end ranges for the program to consider as a uniq hit region e.g. [ [bp_start, bp_end], ... , [bp_start, bp_end] ]
#
#Basic approach:
#1) We gather locations and ranges of hits and separate these based on direction.
#2) We consider overlaping regions of hits in different directions, and chose the one containing an HSP of lowest e-value
#3) We merge overlapping hits and remove duplicates
def segment_hits(hit, evalue_cutoff = 0.001)
  range_info = Struct.new(:evalue, :gene_start, :gene_end, :seg_start, :seg_end, :dir) #holds hit results
  fwd_ranges, rev_ranges = Array.new, Array.new # collection of  direction_specific range_info structs
  fwd_section, rev_section = Array.new, Array.new # holds ranges generated from hit data each as an array [bp_start, bp_end] 
  fwd_evalue, rev_evalue = Array.new, Array.new # corresponding evalues to fwd_section elements
  merged_section = Array.new # holds final merged fwd and rev ranges for both directions (when present)
  
  hit.each do |hsp|
    unless hsp.hit_frame < 0
      fwd_ranges << range_info.new(hsp.evalue, hsp.query_from, hsp.query_to, hsp.hit_from, hsp.hit_to, '==>') if hsp.evalue <= evalue_cutoff
    else
      rev_ranges << range_info.new(hsp.evalue, hsp.query_from, hsp.query_to, hsp.hit_from, hsp.hit_to, '<==') if hsp.evalue <= evalue_cutoff
    end
  end
  
  #Get forward ranges
  this_min, contig_seg_end, ev = 0, 0, 0
  contig_seg_start = 999999999999
  unless fwd_ranges.empty?
    fwd_ranges.each do |r|
      if r.gene_end > this_min
        contig_seg_start = r.seg_start if r.seg_start < contig_seg_start
        contig_seg_end = r.seg_end if r.seg_end > contig_seg_end
      else
        fwd_section << [contig_seg_start, contig_seg_end]
        fwd_evalue << ev
        contig_seg_start = r.seg_start
        contig_seg_end = r.seg_end
        this_min = 0
      end
      this_min = r.gene_end
    end
    fwd_section << [contig_seg_start, contig_seg_end]
    fwd_evalue << ev
  end
  #Get reverse ranges
  this_min, contig_seg_end, ev = 0, 0, 0
  contig_seg_start = 999999999999
  unless rev_ranges.empty?
    rev_ranges.each do |r|
      ev = r.evalue
      if r.gene_end > this_min
        contig_seg_start = r.seg_start if r.seg_start < contig_seg_start
        contig_seg_end = r.seg_end if r.seg_end > contig_seg_end
      else
        rev_section << [contig_seg_start, contig_seg_end]
        rev_evalue << ev
        contig_seg_start = r.seg_start
        contig_seg_end = r.seg_end
        this_min = 0
        end
      this_min = r.gene_end
    end
    rev_section << [contig_seg_start, contig_seg_end]
    rev_evalue << ev
  end
  
  #Save regions in both directions - if both occur in the same range we choose the one with the lower E-value
  if fwd_section.length == 0
    merged_section = rev_section.uniq
  elsif rev_section.length == 0
    merged_section = fwd_section.uniq
  else # we check for overlapping sections and take one with lower e-value
    fwd_section.uniq.each_index do |i|
      rev_section.uniq.each_index do |j|
        if( [ fwd_section[i][0], rev_section[j][0] ].max <= [ fwd_section[i][1], rev_section[j][1] ].min )
          if fwd_evalue[i] <= rev_evalue[j]
            merged_section << [fwd_section[i][0], fwd_section[i][1]]
          else 
            merged_section << [rev_section[j][0], rev_section[j][1]]
          end
        else # both segments don't overlap
          merged_section << [fwd_section[i][0], fwd_section[i][1]]
          merged_section << [rev_section[j][0], rev_section[j][1]]
        end      
      end
    end
  end
  
  #Merge overlapping regions
  merged = true
  while(merged) # continue to loop and check as long as a segment is merged
    merged = false
    for i in 0 ... (merged_section.length - 1)
      for j in (i + 1) ... merged_section.length
        if( [ merged_section[i][0], merged_section[j][0] ].max <= [ merged_section[i][1], merged_section[j][1] ].min )
          merged = true
          low = [merged_section[i][0], merged_section[j][0]].min
          high = [merged_section[i][1], merged_section[j][1]].max
          merged_section[i][0] = low
          merged_section[j][0] = low
          merged_section[i][1] = high
          merged_section[j][1] = high
        end
      end
    end
    merged_section.uniq!
  end   
  return merged_section
end

#========================
#MAIN

#Variables 
all_hits = []
nhits = 0 # hits meeting evalue cutoff
idenhits = 0 # hits meeting query coverage cutoff
match_gene_name = nil # holds regex of name to match
cfinder_seqs = Hash.new # FOR SERVER ONLY
seen_gene_names = Hash.new
print_gene_counts = Hash.new
this_gene_name = ''
hit_info = Struct.new(:gene_name, :contig_num, :full_gene_start, :full_gene_end, :seg_start, :seg_end, :contig_start, :contig_end, :contig_len, :direction) #holds hit results

#Blast User Defaults
evalue_cutoff = 0.001 
gap_align = 'T'
nthreads = 1
expand_hits = true # try to go out from best HSP and expand hit range for other valid hits in same strand direction

#Other User Defaults
query_coverage_cutoff = 30.0 # min query coverage cutoff
cfinder_mode = false # FOR SERVER ONLY - prints CFinderSeqs.fa with seqs with hits if -cfinder flag is provided
filter_gene_name = false # set mode to only print results for a given gene name?
coords = true # return coordinates?
prot_search = false #are queries proteins?
flip = false #print flipped cmap line?
silent = true #print debug output?
adjust_aa2na = false #if the -p flag is set, multiply query coords x 3 to convert back to NA numbering. This assumes no exon splicing is present! 
mul_type = 1 # set to 1 for nucleotide genes, will be set to 3 if protien query flag is thrown

#Print user message and get user arguments - there are some additional options used for debugging hidden by default here
if ARGV.length < 3 || ARGV[0] !~ /\.fn?as?t?a?/i || ARGV[1] !~ /\.fn?as?t?a?/i || ARGV.include?('-help') || ARGV.include?('-h')
  puts "\n================================================================================================"
  puts "SyntenyMapper Vr.#{$VERSION} - Takes FASTA genome/target files and base name for file output and prints\nvalid CMAP lines for use with SyntenyDrawer."
  puts "\nPlease Cite: Veltri, D., Malapi-Wight, M. and Crouch J.A. SimpleSynteny: a web-based tool for"
  puts "visualization of microsynteny across multiple species. Nucleic Acids Research, 44(W1) 2016."
  puts "------------------------------------------------------------------------------------------------"
  puts "Map_Contigs Usage:"
  puts "ruby Map_Contigs_1h.rb <my_query_genes_fasta> <my_subject_contigs_fasta> <output_base_name> [option flags]"
  puts "\nRequired Arguments:"
  puts "<my_query_genes_fasta> :\tFASTA file with gene nucleotide sequences (if proteins use the -p flag) to search for."
  puts "<my_subject_contigs_fasta> :\tBlast Nucleotide Database for organism of interest. Create with makeblastdb."
  puts "<output_base_name> :\t\tBase name (with optional path) to name file output."
  puts "\nOption Flags:"
  puts "-e <num> :\t\tAdjust the E-value cutoff for BLAST (default: #{evalue_cutoff})."
  puts "-n <num> :\t\tPass number of threads to use for BLAST (Default: #{nthreads})"
  puts "-gap <T/F> :\t\tHave BLAST perform gapped alignments? True/False (default: T)."
  puts "-cov <num> :\t\tAdjust the Minimum Query Coverage / Percent Identity for a valid BLAST hit to be shown (default: #{query_coverage_cutoff}, set to 0 to turn off)."
  puts "-f :\t\t\tShow forward and flipped contig directions when printing (default: off)."
  puts "-gene <string> :\tShow ONLY gene results that match '<string>' as gene name (default: off)."
  #puts "-p :\t\t\tForce Protein Mode and use 'tblastn' if <my_query_genes_fasta> contains protein sequences.\n\t\t\tMultiplies coordinates x3 to convert from AA to NA and assumes no exon splicing (default: off).\n\t\t\tCan't use with -p_noadjust option."
  #puts "-p_noadjust :\t\tUse 'tblastn' but DO NOT multiply coordinates x3 to convert from AA to NA (default: off).\n\t\t\tCan't use with -p option."
  #puts "-x :\t\t\tPrint top HSP only, cancel expansion to other valid hits in same strand direction of best HSP (default: off)"
  #puts "-no_coord :\t\tDon't print out the full coordinates, just preview contigs and gene names (default: off).\n\t\t\tCan't use with -f option." 
  puts "-debug :\t\tPrint extra information (default: off)."
  puts "-help :\t\t\tPrint this message.\n\n"
  puts "**This program comes with no warranty**"
  puts "Please see the attached documentation for CMAP file format details."
  puts "Contact Dan Veltri (dan.veltri@gmail.com) if you have problems with the program.\n\n"
  exit
else
  ARGV.each_index do |i|
    if ARGV[i] =~ /-no_coord/i && ARGV.length > 2 then coords = false end
    if ARGV[i] =~ /-p/i && ARGV.length > 2 then prot_search = true; mul_type = 3; adjust_aa2na = true end
    if ARGV[i] =~ /-p_noadjust/i && ARGV.length > 2 then prot_search = true end
    if ARGV[i] =~ /-x/i && ARGV.length > 2 then expand_hits = false end
    if ARGV[i] =~ /(-e)/i && ARGV.length >= (i+1) then !!Float(evalue_cutoff) rescue rasie "Invalid value '#{ARGV[i + 1]}' provided for E-value cutoff!"; evalue_cutoff = ARGV[i + 1].to_f end
    if ARGV[i] =~ /(-cov)/i && ARGV.length >= (i+1) then !!Float(query_coverage_cutoff) rescue rasie "Invalid value '#{ARGV[i + 1]}' provided for query coverage cutoff! Must be a real number 1..100!"; query_coverage_cutoff = ARGV[i + 1].to_f end
    if ARGV[i] =~ /-f/i && ARGV.length > 2 then flip = true end
    if ARGV.length >= (i+1) && ARGV[i] =~ /-gene/i then filter_gene_name = true; match_gene_name = Regexp.new(ARGV[i+1], true) end
    if ARGV.length >= (i+1) && ARGV[i] =~ /-gap/i then if(ARGV[i+1] =~ /fa?l?s?e?/i) then gap_align = 'F' end end
    if ARGV[i] =~ /(-n)/i && ARGV.length >= (i + 1) then !!Float(nthreads) rescue rasie "Invalid value '#{ARGV[i + 1]}' provided for BLAST num_hreads param. Must be a real number >1!"; nthreads = ARGV[i + 1].to_i end
    if ARGV[i] =~ /-de?b?u?g?/i && ARGV.length > 2 then silent = false end
    if ARGV[i] =~ /-cfinder/i && ARGV.length > 2 then cfinder_mode = true end
  end
end

puts "\nWARNING: Only using best HSP for each hit!\n\n" unless silent == true || expand_hits == true

#Check if queries are NA or Protein also see if we have duplicate gene names
query_ff = Bio::FlatFile.auto(ARGV[0])
query_ff.each do |s|
  if seen_gene_names.include?(s.definition) then seen_gene_names[s.definition] += 1 else seen_gene_names[s.definition] = 1; print_gene_counts[s.definition] = 0 end
  if prot_search == false && s.seq =~ /[EFILMPQZ]/i then prot_search = true end
end
if (prot_search == true && silent == false) then puts "Detected a protein query file!" elsif (silent == false) then puts "Detected a nucleotide query file!" end
if (prot_search == false) then print "Blasting (w/#{nthreads}CPU) genes (Query coverage cutoff #{query_coverage_cutoff}; E-value cutoff #{evalue_cutoff}; Gapped Aln: #{gap_align}) in #{ARGV[0]} to #{ARGV[1]}...\n" if silent == false else print "[T]Blasting (w/#{nthreads}CPU) proteins (Quercy coverage cutoff #{query_coverage_cutoff}; E-value cutoff #{evalue_cutoff}; Gapped Aln: #{gap_align}) in #{ARGV[0]} to #{ARGV[1]}...\n" if silent == false end

# Report files for output
blast_report_out = File.open(ARGV[2] + "_with_Query_#{ARGV[0].gsub(/^.*\//,'').gsub(/\.f[astq]+/i,'')}_BLAST_REPORT.csv","w")
skipped_report_out = File.open(ARGV[2] + "_skipped_genes.txt","w") #File to record any genes skipped due to low sequence identity
mask_report_out = File.open(ARGV[2] + "_with_Query_#{ARGV[0].gsub(/^.*\//,'').gsub(/\.f[astq]+/i,'')}_MASKS.txt","w") # File to save BLAST output for user-reporting
blast_report_out.write("qseq_name,sseq_name,alignment_length,num_gaps,qstart,qend,sstart,send,evalue,bitscore\n") #report header

#Generate BLAST DB
`#{$LOCAL_MAKEBLASTDB_LOC} -dbtype nucl -in #{ARGV[1]} -out #{$TOKEN}` unless File.file?("#{$TOKEN}.nin")
puts "Finished making DB! #{$TOKEN}" unless silent == true

if prot_search == true
  blast_report_out.write("TBLASTN Vr. #{$BLAST_VR} RESULTS\n")
  local_blast_factory = Bio::Blast.local('tblastn',"#{$TOKEN}", options="-a #{nthreads} -m 7 -g #{gap_align} -e #{evalue_cutoff}",$LOCAL_BLASTALL_LOC)
else
  blast_report_out.write("BLASTN Vr. #{$BLAST_VR} RESULTS\n")
  local_blast_factory = Bio::Blast.local('blastn',"#{$TOKEN}", options="-a #{nthreads} -m 7 -g #{gap_align} -e #{evalue_cutoff}",$LOCAL_BLASTALL_LOC)
end

#Cycle through queries
query_ff.rewind
query_ff.each do |f|

  next if f.definition.to_s !~ match_gene_name && filter_gene_name == true

  mask_report_out.write("GENE: #{f.definition} - LENGTH: #{f.seq.length}\nSEQ  #{f.seq}\n")
  report = local_blast_factory.query(f.seq)
  query_len = f.seq.length
  best_identity = 0.0
  best_iden_contig = ''
  best_bitscore = 0.0
  valid_hsp_evalue = false
  any_hits = false
  
  #Adjust gene name and see if we've seen it before (append .1, .2 etc. for repeats)
  if seen_gene_names[f.definition] > 1
    print_gene_counts[f.definition] += 1
    this_gene_name = f.definition.gsub(/\s+/,'_') + ".#{print_gene_counts[f.definition]}"
  else
    print_gene_counts[f.definition] = 1
    this_gene_name = f.definition.gsub(/\s+/,'_') 
  end

  puts "\n-------------------------\n#{this_gene_name}" unless silent == true
  if prot_search == true
    puts "Query Length: #{f.seq.length}aa" unless silent == true
  else
    puts "Query Length: #{f.seq.length}bp" unless silent == true
  end
  
  report.each do |hit|
    hit_def = hit.target_def
    mask_report_out.write("CONTIG: #{hit_def}\n")
    puts "\nCONTIG: #{hit_def}\n" unless silent == true

    hit_search_segments = segment_hits(hit, evalue_cutoff) # contains 2d array of ranges to search, [  [bp_start, bp_end] , [bp_start, bp_end] , ... ]
    
    hit_search_segments.each_index do |sidx|
      dirset = false
      firsthsp_only = false
      hitdir = ''
      evals = Array.new
      lowest_subj = 0
      highest_subj = 0
      lowest_query = 0
      highest_query = 0
      fullquery_iden = 0.0

      #Figure identity of all tiled HSP's across full-length query and locate anchor segment (best scoring HSP)
      query_bit_vector = "0" * query_len #bit-vector same length of query to keep track of tiles
      mask_report_out.write("INIT #{query_bit_vector}\n")

      hit.each do |hsp|
        blast_report_out.write("#{f.definition},#{hit.target_def},#{hsp.align_len},#{hsp.gaps},#{hsp.query_from},#{hsp.query_to},#{hsp.hit_from},#{hsp.hit_to},#{hsp.evalue},#{hsp.bit_score}\n") #record hit for report
        
        if hsp.evalue <= evalue_cutoff && firsthsp_only == false && ( [hsp.hit_from, hit_search_segments[sidx][0]].max <= [hsp.hit_to, hit_search_segments[sidx][1]].min )

          #Lock anchor and hit direction to best/initial HSP
          if dirset == false
            best_bitscore = hsp.bit_score

            lowest_query = [hsp.query_from,hsp.query_to].min
            highest_query = [hsp.query_from,hsp.query_to].max
            lowest_subj = [hsp.hit_from,hsp.hit_to].min
            highest_subj = [hsp.hit_from,hsp.hit_to].max
          
            #puts hsp.hit_frame
            unless hsp.hit_frame < 0
              hitdir = '==>'
            else
              hitdir = '<=='
            end
            puts "Gene Direction of Best HSP: #{hitdir}" unless silent == true
            puts "Best HSP Ranges - Query: [#{lowest_query},#{highest_query}] Subj: [#{lowest_subj},#{highest_subj}]" unless silent == true
            valid_hsp_evalue = true #signal at least one valid hsp found
            dirset = true
          end


          #Only consider hits in same direction as initial/best HSP 
          if (hsp.hit_frame < 0 && hitdir == '<==') || (hsp.hit_frame >= 0 && hitdir == '==>')

            this_lowest_query = [hsp.query_from,hsp.query_to].min
            this_highest_query = [hsp.query_from,hsp.query_to].max
            this_lowest_subj = [hsp.hit_from,hsp.hit_to].min
            this_highest_subj = [hsp.hit_from,hsp.hit_to].max
            puts "This HSP - Query: [#{this_lowest_query},#{this_highest_query}] Subj: [#{this_lowest_subj},#{this_highest_subj}]" unless silent == true
            
            #Check for valid match
            pass_check = false
            if ((this_lowest_query <= lowest_query && this_lowest_subj <= lowest_subj && this_highest_subj <= highest_subj && this_lowest_subj >= (lowest_subj - 2 * query_len)) || (this_highest_query >= highest_query && this_highest_subj >= highest_subj && this_lowest_subj >= lowest_subj && this_highest_subj <= (highest_subj + 2 * query_len))) && (hitdir == '==>')
              pass_check = true      
            elsif ((this_lowest_query <= lowest_query && this_lowest_subj >= lowest_subj && this_highest_subj >= highest_subj && this_lowest_subj <= (lowest_subj + 2 * query_len)) || (this_highest_query >= highest_query && this_highest_subj <= highest_subj && this_lowest_subj <= lowest_subj && this_highest_subj >= (highest_subj - 2 * query_len))) && (hitdir == '<==')
              pass_check = true        
            end
            if(pass_check == true)
              lowest_query = [this_lowest_query,lowest_query].min
              highest_query = [this_highest_query,highest_query].max
              lowest_subj = [this_lowest_subj,lowest_subj].min
              highest_subj = [this_highest_subj,highest_subj].max

              evals << hsp.evalue
              full_mask = "0" * query_len #initialize full-size mask for this HSP to query length

              #Create mask for just HSP segment, skip over query gaps, match=1, mismatch=0
              qry_mask = ''
              addgap = 0
              puts "HSP: #{hsp.qseq.gsub(/\-/,'')}" unless silent == true
              for i in 0 ... hsp.qseq.length

                if hsp.qseq[i,1] == '-'
                  addgap += 1
                elsif hsp.qseq[i,1] == 'X'
                  qry_mask += '0'
                else
                  qry_mask += '1'
                end
              end
            
              if hitdir == '<==' && prot_search == false
                cleaned_hsp_qseq = Bio::Sequence.auto(hsp.qseq.gsub(/\-/,'')).reverse_complement
              else
                cleaned_hsp_qseq = hsp.qseq.gsub(/\-/,'')
              end

              hsp_mask = " " * query_len
              hsp_mask[(this_lowest_query - 1),  cleaned_hsp_qseq.length] = cleaned_hsp_qseq
            
              full_mask[(this_lowest_query - 1), cleaned_hsp_qseq.length] = qry_mask #splice HSP segment mask into full-length mask
              query_bit_vector = (query_bit_vector.to_i(2) | full_mask.to_i(2)).to_s(2).rjust(query_len, "0") #bitwise OR with current query_bit_vector state 
              mask_report_out.write("\nMASK #{full_mask}\nHSP  #{hsp_mask} - LENGTH:#{qry_mask.length}, EVALUE:#{hsp.evalue}\n")
            end
          end

          firsthsp_only = true if expand_hits == false #if we're not expanding from first HSP, stop coverage calculation now
        end
      end

      nhits += 1 if valid_hsp_evalue == true

      puts "Range of All Valid HSPs - Query: [#{lowest_query},#{highest_query}] Subj: [#{lowest_subj},#{highest_subj}]" unless silent == true

      fullquery_iden = (query_bit_vector.count('1') / query_len.to_f) * 100.0
      puts "Query Coverage/Identity: #{fullquery_iden}% (ignores X's and gaps)" unless silent == true
      mask_report_out.write("\nFNL  #{query_bit_vector} \nIDENTITY: #{fullquery_iden}\n\n")    

      if best_identity < fullquery_iden
        best_identity = fullquery_iden
        best_iden_contig = hit_def
      end

      if fullquery_iden >= query_coverage_cutoff && valid_hsp_evalue == true
        any_hits = true
        idenhits += 1
        this_contig_seq, this_contig_seq_revcomp, this_contig_len = get_contig(ARGV[1], hit_def)

        if adjust_aa2na == true
          diff_high = (query_len * 3) - ((query_len - highest_query) * 3)
          diff_low = ((lowest_query - 1) * 3) + 1
          all_hits << hit_info.new(this_gene_name, hit_def, 1, (query_len * 3), diff_low, diff_high, lowest_subj, highest_subj, this_contig_len, hitdir)
        else
          all_hits << hit_info.new(this_gene_name, hit_def, 1, query_len, lowest_query, highest_query, lowest_subj, highest_subj, this_contig_len, hitdir)
        end

        # SERVER ONLY - CFinder mode
        if(cfinder_mode == true && cfinder_seqs.has_key?(hit_def) == false)
          cfinder_seqs[hit_def] = this_contig_seq 
        end

      end
    end
  end

  #record if we didn't get any matches for this query
  if any_hits == false 
    if best_identity < query_coverage_cutoff && valid_hsp_evalue == true#there was an HSP but didn't meet identity cutoff
      skipped_report_out.write("#{this_gene_name}: No matches satisfying full-length sequence coverage cutoff: #{query_coverage_cutoff}% (highest found: #{"%0.2f" % best_identity}%)\n") 
      puts "\tHit(s) on contigs missed sequence coverage cutoff: #{"%0.3f" % best_identity} (Cutoff:#{query_coverage_cutoff})" unless silent == true

    else #gene didn't map due to no significant HSP's at this e-value
      skipped_report_out.write("#{this_gene_name}: No significant matches found at E-value cutoff: #{evalue_cutoff}\n") 
      puts "\tHit(s) on all contigs did not satisfy sequence e-value cutoff: #{"%0.3f" % evalue_cutoff}" unless silent == true
    end
  end
  
end

# SERVER ONLY - Save contig hits if CFinder mode is on
if cfinder_mode == true
  puts "\nWriting contigs with hits out to #{ARGV[2]}_CFinder_Hits.fa for CFinder..." unless silent == true
  puts cfinder_seqs unless silent == true
  cfinder_out = File.open(ARGV[2] + "_CFinder_Hits.fa","w")
  cfinder_seqs.each {|k,v| cfinder_out.write(">#{k}\n#{v}\n") }
  cfinder_out.close
end

#Close reports
skipped_report_out.close
blast_report_out.close
mask_report_out.close

#Prep CMAP output files
fwd_out = File.open(ARGV[2] + "_fwd.cmap","w")
rev_out = File.open(ARGV[2] + "_rev.cmap","w") if flip == true

#Make sure we got a hit based on e-value and identity 
if idenhits == 0
  puts "\nNo Hits Found Satisfying Coverage Cutoff: #{"%0.2f" % query_coverage_cutoff}%"
  fwd_out.write("\nNo Hits Found Satisfying Coverage Cutoff: #{"%0.2f" % query_coverage_cutoff}%")
  rev_out.write("\nNo Hits Found Satisfying Coverage Cutoff: #{"%0.2f" % query_coverage_cutoff}%") if flip == true
  fwd_out.close()
  rev_out.close() if flip == true
  exit
elsif nhits == 0
  puts "\nNo Hits Found at E-value: #{"%0.4f" % evalue_cutoff}"
  fwd_out.write("\nNo Hits Found at E-value: #{"%0.4f" % evalue_cutoff}")
  rev_out.write("\nNo Hits Found at E-value: #{"%0.4f" % evalue_cutoff}") if flip == true
  fwd_out.close()
  rev_out.close() if flip == true
  exit
else
  print "\n DONE! (#{nhits} hits found)\n\nGenerating CMAP line(s):\n\n" unless silent == true
end

#Print CMAPs of hits in order of original genome file
index_order = []
contigs = all_hits.uniq.collect {|h| h.contig_num }
Bio::FlatFile.auto(ARGV[1]).each {|f| res = contigs.find_index(f.definition); index_order << res unless res == nil} 
index_order.each do |idx|
  contig = contigs[idx]
  cgenes = all_hits.select{|c| c.contig_num == contig}
  cgenes.sort! {|a,b| a.contig_start <=> b.contig_start}
  last_name = ''
  print_string = '{1 '
  cgenes.each do |g|

    if g.direction == '<=='
      print_gene_format = "#{g.direction} #{g.gene_name}"
    elsif g.direction == '==>'
      print_gene_format = "#{g.gene_name} #{g.direction}"
    else
      raise "ERROR IN DIRECTION: #{g}"
      exit
    end
   
    if coords == true
        print_string += " [ #{[(g.contig_start - (g.seg_start - g.full_gene_start)),1].max}|#{g.contig_start} (#{([(g.contig_start - g.seg_start - 1),1].min).abs}|#{g.seg_start}  #{print_gene_format}  #{g.seg_end}|#{g.full_gene_end}) #{g.contig_end}|#{[(g.contig_end + (g.full_gene_end - g.seg_end)),g.contig_len].min} ] "
    else
      print_string += "[ #{print_gene_format} ]" unless g.gene_name == last_name
      last_name = g.gene_name
    end
  end
  
  print_string += " #{cgenes[0].contig_len}}\n"
  
  if flip == true
    puts "Fwd: #{contig} " + print_string
    fwd_out.write("#{contig} #{print_string}")
    puts "\nRev: #{contig} " + flip_contig(print_string) + "\n\n"
    rev_out.write("#{contig} " + flip_contig(print_string))
  else
    puts "#{contig} " + print_string
    fwd_out.write("#{contig} #{print_string}")
  end
end
fwd_out.close
rev_out.close if flip == true

#Cleanup BLAST DB Files
Dir.glob("#{$TOKEN}.n*").each {|f| File.delete(f)} if $CLEAN_DB == true

#========================
#END PROGRAM
#========================
